# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 22 August 2022

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import ipywidgets as widgets
import copy as cp
import pickle as pkl

from .base import widget_base
from .baseline import baseline
from .fit import fit
from .int_peaks import int_peaks
from .labels import labels
from .display_tools import display_tools
from .export import export

from scipy import integrate as scpyintegrate
import functools

class main_menu(widget_base):
    def __init__(self,data,load_file=None,fig=None,ax=None,exp_path=''):
        '''
        Initializes a widget instance
        '''
        # Close any earlier plots
        plt.close()

        # Initialize widget parameters
        self.widget_on = True
        self.active_widget = self
        self.prev_widget = self
        self.menu,self.sub_menu,self.info_menu = widgets.HBox(),widgets.HBox(),widgets.HBox()

        # Check for an load dictionary
        if load_file is not None:
            plt.ioff()
            load = cp.deepcopy(pkl.load(open(load_file,'rb')))
            fig = load['fig']; ax = fig.axes
            plt.ion()
        else: load = None

        # Initialize figure
        self.fig_init(fig,ax)

        if load is None:
            # Initialize artist list
            self.artists = dict.fromkeys(self.fig.axes,{})
            # Import the data from the initialized instance
            self.artists[self.ax]['Input Data'],self.input_data = self.setup_data(data)
            self.artists['Primary Artists'] = {self.ax:'Input Data'}
            self.artists['Interactive Axes'] = [self.ax]

            # Initialize data list
            self.data = dict.fromkeys(self.fig.axes,{})

        else:
            self.artists = load['artists_global']
            self.data = load['data_global']

        # Initialize all sub_menus
        self.baseline = baseline(self.fig,menu=self.sub_menu,artists_global=self.artists,data_global=self.data,load=load)
        self.fit = fit(self.fig,menu=self.sub_menu,param_menu=self.info_menu,artists_global=self.artists,data_global=self.data,load=load)
        self.int_peaks = int_peaks(self.fig,menu=self.sub_menu,artists_global=self.artists,data_global=self.data,load=load)
        self.labels = labels(self.fig,menu=self.sub_menu,artists_global=self.artists,data_global=self.data,load=load)
        self.display_tools = display_tools(self.fig,menu=self.sub_menu,artists_global=self.artists)
        self.export = export(main_menu=self,menu=self.sub_menu,exp_path=exp_path)

        # Widget list
        self.widgets = {
        'baseline':self.baseline,
        'fit':self.fit,
        'int_peaks':self.int_peaks,
        'labels':self.labels,
        'display_tools':self.display_tools,
        'export':self.export
        }

        # Setup the main menu and a blank sub_menu
        self.button_list = self.setup_buttons()
        self.place_menu(self.menu,self.button_list)
        self.place_menu(self.sub_menu,[])
        self.place_menu(self.info_menu,[])
        self.fig_show(self.fig)
        plt.show()

    # Define all main menu buttons and their functionality
    def setup_buttons(self):

        ##########################################
        ## ON OFF BUTTON
        ##########################################

        # Add an on/off toggle to activate or deactive the widget
        self.on_off_button = widgets.ToggleButton(description='On',value=True,layout = widgets.Layout(width='45px'))

        def on_on_off_button_clicked(b, self = self):

            # If the widget is on, turn it off
            if self.widget_on:
                # Set the description
                self.on_off_button.description = 'Off'
                # Turn off the main widget
                self.deactivate(main=True)
                # Turn off all buttons in the main menu
                for tb in self.button_list:
                    # Turn it off if except for the selected menu
                    if tb != self.on_off_button: tb.value = False
                # Deactivate the current active widget
                if self.active_widget != self: self.active_widget.deactivate()
            # If the widget is off, turn it on
            else:
                self.on_off_button.description = 'On'
                self.activate()

        self.on_off_button.observe(on_on_off_button_clicked)

        ##########################################
        ## BUTTON TEMPLATE
        ##########################################

        def menu_button(self,menuclass,button):
            '''
            Generic button function for all new buttons
            '''
            def func(b,self=self):
                if not self.widget_on:
                    for tb in self.button_list:
                        if tb != self.on_off_button: tb.value = False
                    return
                # If the widget is on, turn it off
                if menuclass.widget_on:
                    # Turn off the widget
                    menuclass.deactivate()
                    # Set the previous widget
                    self.prev_widget = menuclass
                    # Set the active widget to the main menu
                    self.active_widget = self
                # If the widget is off, turn it on
                else:
                    # For each button in the main menu
                    for tb in self.button_list:
                        # Turn it off if except for the selected menu
                        if tb != button and tb != self.on_off_button: tb.value = False
                    # Set the active widget to the selected menu
                    self.active_widget = menuclass
                    # If moving to the export menu
                    if self.active_widget == self.display_tools: self.display_tools.update(self.artists,None)
                    # Update the menu
                    self.prev_widget.update(self.artists,self.data)
                    # Turn off all the menu's buttons
                    for tb in menuclass.toggle_buttons: tb.value = False
                    # Turn on the selected menu
                    menuclass.activate()

                    # Listen for a click
                    self.listen(menuclass)
            return func

        ##########################################
        ## ADD WIDGETS
        ##########################################

        # Baseline toggle
        self.baseline_button = widgets.ToggleButton(description='Baseline',value=False)
        self.baseline_button.observe(menu_button(self,self.baseline,self.baseline_button), names=['value'])

        ##########################################

        # Peaks toggle
        self.int_peaks_button = widgets.ToggleButton(description='Integrate/Peaks',value=False)
        self.int_peaks_button.observe(menu_button(self,self.int_peaks,self.int_peaks_button), names=['value'])

        # ##########################################

        # Integrate toggle
        self.fit_button = widgets.ToggleButton(description='Fit',value=False)
        self.fit_button.observe(menu_button(self,self.fit,self.fit_button), names=['value'])

        # ##########################################

        # Labels toggle
        self.labels_button = widgets.ToggleButton(description='Labels',value=False)
        self.labels_button.observe(menu_button(self,self.labels,self.labels_button), names=['value'])

        # ##########################################
        #
        # Display toggle
        self.display_tools_button = widgets.ToggleButton(description='Display',value=False)
        self.display_tools_button.observe(menu_button(self,self.display_tools,self.display_tools_button), names=['value'])

        # ##########################################
        #
        # Export toggle
        self.export_button = widgets.ToggleButton(description='Export',value=False)
        self.export_button.observe(menu_button(self,self.export,self.export_button), names=['value'])

        ##########################################
        ## BUTTON LIST
        ##########################################

        return [self.on_off_button,self.baseline_button,self.int_peaks_button,self.fit_button,self.labels_button,self.display_tools_button,self.export_button]

    def setup_data(self,data):
        # Setup data
        data_dict = {}
        if isinstance(data,mpl.lines.Line2D):
            data_dict['xpoints'],data_dict['ypoints'] = data.get_data()
            artist = data
        elif isinstance(data,list) or isinstance(data,tuple) or isinstance(data,np.ndarray):
            if isinstance(data[0],mpl.lines.Line2D):
                data_dict['xpoints'],data_dict['ypoints'] = data[0].get_data()
                artist = data[0]
            elif len(data) == 2:
                data_dict['xpoints'],data_dict['ypoints'] = data
                artist = self.ax.plot(data_dict['xpoints'],data_dict['ypoints'])[0]
        return artist,data_dict

    # Initialize all the default colors and color targets
    def color_init(self):
        self.color_target = 'Marker - Primary'
        self.primary_color = 'deepskyblue'
        self.new_color = 'red'
        self.secondary_color = 'green'
        self.fit_color = 'darkviolet'
        self.bscolor = 'blue'
        self.fill_color = 'skyblue'
        self.marker = '*'

    def data_tree(self,*kwargs):
        self.active_widget.update(self.artists,self.data)
        ddict = self.get_data()
        for arg in kwargs:
            if arg in ddict.keys():
                ddict = ddict[arg]
        self.printdict(ddict)

    def artist_tree(self,*kwargs):
        self.active_widget.update(self.artists,self.data)
        adict = self.get_artists()
        for arg in kwargs:
            if arg in adict.keys():
                adict = adict[arg]
        self.printdict(adict)

    def get_data(self):
        # Update the data
        self.active_widget.update(self.artists,self.data)
        # Get the list of axes
        axes = self.get_axes()
        # Create a new data dictionary for data export
        exp_data = {}
        # Fill the new dictionary
        for ind,ax in enumerate(axes):
            exp_data[f'Axis {ind+1}'] = self.data[ax]
        # Return the export dictionary
        return exp_data

    def get_artists(self):
        # Update the artists
        self.active_widget.update(self.artists,self.data)
        # Get the list of axes
        axes = self.get_axes()
        # Create a new data dictionary for data export
        exp_artists = {}
        # Fill the new dictionary
        for ind,ax in enumerate(axes):
            exp_artists[f'Axis {ind+1}'] = self.artists[ax]
        # Return the export dictionary
        return exp_artists

    def export_dict(self):
        self.active_widget.update(self.artists,self.data)
        export = {
        'artists_global': self.artists,
        'data_global': self.data,
        }
        for key in self.widgets.keys():
            if key == 'export' or key == 'display_tools':continue
            w_export = {
            'data': self.widgets[key].get_data(),
            'artists': self.widgets[key].get_artists(),
            'style': self.widgets[key].style,
            'info': self.widgets[key].info
            }
            export[key] = w_export

        export['fig'] = self.fig

        return export

'''

'''



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 23 September 2022

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools

from base import widget_base

class display_tools(widget_base):
    def __init__(self,fig,menu=None,artists_global=None):
        '''
        '''
        self.widget_on = False
        self.fig = fig

        # Initialize all global matplotlib artists
        self.artists_global = self.pull_artists(artists_global)

        # Initialize defaults
        self.info = self.setup_defaults()

        # Initialize all buttons
        self.button_list,self.toggle_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu

    def setup_defaults(self):
        info = {
        'axis':self.fig.axes[0],
        'property':'',
        'property_value':None,
        'artist':self.artists_global[self.fig.axes[0]]['Input Data'],
        'interactive_mode':'off',
        'active_ax':None,
        'scale':'linear'
        }

        return info

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.axis_select = widgets.Select(description='Axis:',options=range(1,len(self.fig.axes)+1),rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='95px'))

        def on_axis_selected(b, self = self):

            self.info['axis'] = self.fig.axes[self.axis_select.value-1]
            self.artist_select.options = list(self.artists_global[self.info['axis']].keys())

        self.axis_select.observe(functools.partial(on_axis_selected, self=self))

        #####################

        # Add a button widget to add new points
        self.artist_select = widgets.Select(description='Artist:',options=list(self.artists_global[self.info['axis']].keys()),rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='210px'))

        def on_artist_selected(b, self = self):

            if self.artist_select.value in self.artists_global[self.info['axis']].keys():
                self.info['artist'] = self.artists_global[self.info['axis']][self.artist_select.value]
                artist = self.info['artist']
            else: return

            # If the artist is a plot
            if isinstance(artist,matplotlib.lines.Line2D):
                self.property_select.options = ['Color','Marker','Markersize','Style','Width','Alpha']
            # If the artist is a vlines collections
            elif isinstance(artist,matplotlib.collections.LineCollection):
                self.property_select.options = ['Color','Style','Width','Alpha']
            # If the artist is a fill_between
            elif isinstance(artist,matplotlib.collections.PolyCollection):
                self.property_select.options = ['Color','Alpha']

        self.artist_select.observe(functools.partial(on_artist_selected, self=self))

        #####################

        # Add a button widget to add new points
        self.property_select = widgets.Select(description='Property:',options=['Color','Marker','Markersize','Style','Width','Alpha'],rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='180px'))

        def on_property_selected(b, self = self):

            # Change saved property value
            self.info['property'] = self.property_select.value
            artist = self.info['artist']

            if self.info['property'] == 'Color':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.color_picker,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                if isinstance(artist,matplotlib.collections.PolyCollection) or isinstance(artist,matplotlib.collections.LineCollection):
                    color = artist.get_facecolor()
                    hexcolor = '#{:02x}{:02x}{:02x}'.format(*[int(n) for n in color[0][:3]*255])
                    self.color_picker.value = hexcolor
                else: self.color_picker.value = artist.get_color()
            elif self.info['property'] == 'Marker':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.prop_text,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                self.prop_text.value = artist.get_marker()
                self.prop_text.description = 'Marker:'
            elif self.info['property'] == 'Markersize':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.prop_float,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                self.prop_float.value = artist.get_markersize()
                self.prop_float.description = 'Marker Size:'
            elif self.info['property'] == 'Style':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.prop_text,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                self.prop_text.value = artist.get_linestyle()
                self.prop_text.description = 'Style:'
            elif self.info['property'] == 'Width':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.prop_float,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                self.prop_float.value = artist.get_linewidth()
                self.prop_float.description = 'Width:'
            elif self.info['property'] == 'Alpha':
                self.button_list = [self.axis_select,self.artist_select,self.property_select,self.prop_float,self.scale_button,self.scale_select]
                self.place_menu(self.menu,self.button_list)
                if artist.get_alpha() is None: self.prop_float.value = 1
                else: self.prop_float.value = artist.get_alpha()
                self.prop_float.description = 'Alpha:'

        self.property_select.observe(functools.partial(on_property_selected, self=self))

        #####################

        # Add a button widget to add new points
        self.color_picker = widgets.ColorPicker(description='Color:',value='deepskyblue',concise=True,style={'description_width': 'initial'},layout = widgets.Layout(width='90px'))

        def on_color_picked(b, self = self):

            artist = self.info['artist']

            artist.set(color=self.color_picker.value)

            plt.show()

        self.color_picker.observe(functools.partial(on_color_picked, self=self))

        #####################

        # Add a button widget to add new points
        self.prop_text = widgets.Text(description='Text:',value='',style={'description_width': 'initial'},layout=widgets.Layout(width='150px'))

        def on_prop_text_entered(b, self = self):

            artist = self.info['artist']

            if self.info['property'] == 'Style':
                if self.prop_text.value not in ['-', '--', '-.', ':', 'None', ' ', '', 'solid', 'dashed', 'dashdot', 'dotted']:
                    artist.set(linestyle='-')
                    return
                artist.set(linestyle=self.prop_text.value)
            if self.info['property'] == 'Marker':
                if self.prop_text.value not in np.concatenate(list(matplotlib.lines.Line2D.markers.items())): return
                artist.set(marker=self.prop_text.value)

            plt.show()

        self.prop_text.observe(functools.partial(on_prop_text_entered, self=self))

        #####################

        # Add a button widget to add new points
        self.prop_float = widgets.FloatText(description='Float:',value=1,style={'description_width': 'initial'},layout=widgets.Layout(width='150px'))

        def on_prop_float_entered(b, self = self):

            artist = self.info['artist']

            if self.info['property'] == 'Markersize':
                artist.set(markersize=self.prop_float.value)
            if self.info['property'] == 'Width':
                artist.set(linewidth=self.prop_float.value)
            if self.info['property'] == 'Alpha':
                if self.prop_float.value < 0 or self.prop_float.value > 1:
                    artist.set(alpha=1)
                    return
                artist.set(alpha=self.prop_float.value)

            plt.show()

        self.prop_float.observe(functools.partial(on_prop_float_entered, self=self))

        #####################

        # Add a button widget to save the figure
        self.scale_button = widgets.ToggleButton(description='Scale Plot')

        def on_scale_button_clicked(b, self = self):

            # Turn on add_range mode
            if self.info['interactive_mode'] == 'scale':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'scale'

        self.scale_button.observe(functools.partial(on_scale_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.scale_select = widgets.Select(description='Scale:',options=['linear','log','symlog','logit'],rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='140px'))

        def on_scale_selected(b, self = self):

            self.info['scale'] = self.scale_select.value

        self.scale_select.observe(functools.partial(on_scale_selected, self=self))

        #####################

        return [self.axis_select,self.artist_select,self.property_select,self.color_picker,self.scale_button,self.scale_select],[self.scale_button]

    def __call__(self,event):

        if not self.widget_on: return
        self.info['active_ax'] = None

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        for ax in self.fig.axes:
            # Determine which axis was clicked on
            if event.inaxes == ax: self.info['active_ax'] = ax
            # If the axis is not interactive, do not add new artists
            if ax not in self.artists_global['Interactive Axes']: continue

        # If a click is not within the axis, do nothing
        if self.info['active_ax'] == None: return
        else: ax = self.info['active_ax']

        ##########################################
        ## ADD RANGE
        ##########################################

        # If in add range mode
        if self.info['interactive_mode'] == 'scale':
            ax.set_yscale(self.info['scale'])

        plt.show()

    def update(self,artists,data):
        self.artists_global = artists
        self.axis_select.options = range(1,len(self.fig.axes)+1)
        self.artist_select.options = list(self.artists_global[self.info['axis']].keys())



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 14 September 2022

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools

from scipy.signal import find_peaks,medfilt
from scipy.integrate import simps
import seaborn as sns

from .base import widget_base

class int_peaks(widget_base):
    def __init__(self,fig,menu=None,data=None,artists_global=None,data_global=None,load=None):
        '''
        '''
        self.widget_on = False
        # Setup the figure
        self.fig = fig

        # Initialize all global matplotlib artists
        self.artists_global = self.pull_artists(artists_global)
        # Initialize all global data
        self.data_global = self.pull_data(data_global)
        # Create a dictionary using the axes as keys
        if load is None: self.artists = {}
        else:
            # If load is not None, subscript it to the load int_peaks dict
            load = load['int_peaks']
            self.artists = load['artists']

        # Initialize defaults
        self.style,self.info,self.data = self.setup_defaults(load)

        # Initialize all buttons
        self.button_list,self.toggle_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu

    def setup_defaults(self,load=None):

        if load is not None:
            load['info']['interactive_mode'] = 'off'
            return load['style'],load['info'],load['data']

        style = {
        'fill_color':'deepskyblue',
        'alpha':0.6,
        'vline_style':'--',
        'vline_width':2,
        'color_palette':'Set2',
        'peak_style':'-',
        'peak_width':1
        }

        info = {
        # Initialize the add, move, and delete
        'interactive_mode':'off',
        # Initialize an active axis
        'active_ax':None,
        # Initialize a selected boolean
        'selected':False,
        # Set an interactive click distance
        'click_dist':0.02,
        # Initialize a height
        'height':None,
        # Initialize a distance
        'distance':None
        }

        data = dict.fromkeys(self.artists_global['Interactive Axes'],self.data_init())

        return style,info,data

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.add_range_button = widgets.ToggleButton(description='Add Range')

        def on_add_range_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.add_range_button: button.value = False

            # Turn on add_range mode
            if self.info['interactive_mode'] == 'add_range':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'add_range'

        self.add_range_button.observe(functools.partial(on_add_range_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.adjust_range_button = widgets.ToggleButton(description='Adjust Range')

        def on_adjust_range_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.adjust_range_button: button.value = False

            # Turn on adjust_range mode
            if self.info['interactive_mode'] == 'adjust_range':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'adjust_range'

        self.adjust_range_button.observe(functools.partial(on_adjust_range_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.integrate_button = widgets.ToggleButton(description='Integrate')

        def integrate_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.integrate_button: button.value = False

            # Turn on integrate mode
            if self.info['interactive_mode'] == 'integrate':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'integrate'

        self.integrate_button.observe(functools.partial(integrate_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.find_peaks_button = widgets.ToggleButton(description='Find Peaks')

        def find_peaks_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.find_peaks_button: button.value = False

            # Turn on find peaks mode
            if self.info['interactive_mode'] == 'find_peaks':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'find_peaks'

        self.find_peaks_button.observe(functools.partial(find_peaks_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.height_floattext = widgets.FloatText(description='Height:',value=0,style={'description_width':'initial'},layout = widgets.Layout(width='115px'))

        def height_floattext_clicked(b, self = self):

            if self.height_floattext.value != 0:
                self.info['height'] = self.height_floattext.value
            else: self.info['height'] = None

        self.height_floattext.observe(functools.partial(height_floattext_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.distance_floattext = widgets.FloatText(description='Distance:',value=0,style={'description_width':'initial'},layout=widgets.Layout(width='130px'))

        def distance_floattext_clicked(b, self = self):

            if self.distance_floattext.value != 0:
                self.info['distance'] = self.distance_floattext.value
            else: self.info['distance'] = None

        self.distance_floattext.observe(functools.partial(distance_floattext_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.clear_axis_button = widgets.ToggleButton(description='Clear Plot')

        def clear_axis_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.clear_axis_button: button.value = False

            # Turn on clear mode
            if self.info['interactive_mode'] == 'clear':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'clear'

        self.clear_axis_button.observe(functools.partial(clear_axis_button_clicked, self=self))

        #####################

        return [self.add_range_button,self.adjust_range_button,self.integrate_button,self.find_peaks_button,self.height_floattext,self.distance_floattext,self.clear_axis_button],[self.add_range_button,self.adjust_range_button,self.integrate_button,self.find_peaks_button,self.clear_axis_button]

    ##########################################
    ## UPDATE METHODS
    ##########################################

    # Method to create a new artist list
    def new_artists(self,ax,style):
        # Find the y axis minimum and maximum
        ymin,ymax = ax.get_ylim()

        artists = {
        'Bounds': ax.vlines([],ymin,ymax,linestyle=self.style['vline_style'],linewidth=self.style['vline_width']),
        'New Bound': ax.vlines([],ymin,ymax,linestyle=self.style['vline_style'],linewidth=self.style['vline_width'],color='red'),
        'Selected Bound': ax.vlines([],ymin,ymax,linestyle=self.style['vline_style'],linewidth=self.style['vline_width'],color='green'),
        'Peaks': ax.vlines([],ymin,ymax,linestyle=self.style['peak_style'],linewidth=self.style['peak_width'],color='red'),
        'Integration': ax.fill_between([],[])
        }
        return artists

    def data_init(self):
        return {
        'Range':[],
        'Integration':{'xdata':[],'ydata':[],'Display':False},
        'Peaks':{'xdata':[],'Display':False},
        }

    ##########################################
    ## EVENT HANDLER
    ##########################################

    def __call__(self,event):

        if not self.widget_on: return
        self.info['active_ax'] = None

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        for ax in self.fig.axes:
            # Determine which axis was clicked on
            if event.inaxes == ax: self.info['active_ax'] = ax
            # If the axis is not interactive, do not add new artists
            if ax not in self.artists_global['Interactive Axes']: continue
            # If the artist is interactive, initialize the appropriate artists
            self.update_dictionaries(ax,artist_key='Bounds',data_key='Peaks')
            self.update(self.artists_global,self.data_global)

        # If a click is not within the axis, do nothing
        if self.info['active_ax'] == None: return
        else: ax = self.info['active_ax']

        ##########################################
        ## ADD RANGE
        ##########################################

        # If in add range mode
        if self.info['interactive_mode'] == 'add_range':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return
            # Check to see how many bounds are already in the axis
            if len(self.data[ax]['Range']) == 2: return

            # Clear all other points
            self.clear(self.artists[ax]['New Bound'],ax=ax,vlines=True)
            self.clear(self.artists[ax]['Selected Bound'],ax=ax,vlines=True)
            self.info['selected'] = False

            # Add the point to the array of boundaries
            self.data[ax]['Range'].append(event.xdata)

            # Replot all the points, including the extra one
            self.set_segments(self.artists[ax]['Bounds'],self.data[ax]['Range'],ax)

        ##########################################
        ## ADJUST RANGE
        ##########################################

        elif self.info['interactive_mode'] == 'adjust_range':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return

            # If a boundary is not selected
            if not self.info['selected']:

                # If there was a new boundary, clear it
                self.clear(self.artists[ax]['New Bound'],ax=ax,vlines=True)
                # Find the distance between the click and each boundary
                self.dat = event.xdata
                hdists = self.hdist(event.xdata,self.data[ax]['Range'],ax)
                self.hdists = hdists
                # Find the index of the point closest to the click
                self.info['close_bound'] = np.nanargmin(hdists)

                # If the point is close to the click
                if hdists[self.info['close_bound']] < self.info['click_dist']:

                    # Replot the selected boundary in a different color
                    self.set_segments(self.artists[ax]['Selected Bound'],[self.data[ax]['Range'][self.info['close_bound']]],ax)
                    # State that a point has been selected
                    self.info['selected'] = True

            # If a point has already been selected
            else:
                # Remove the bound from the data array
                self.data[ax]['Range'] = np.delete(self.data[ax]['Range'],self.info['close_bound'])
                # Remove the temporary plotted point
                self.clear(self.artists[ax]['Selected Bound'],ax=ax,vlines=True)

                # Plot the new line
                self.set_segments(self.artists[ax]['New Bound'],[event.xdata],ax)
                # Add the point to the array of points
                self.data[ax]['Range'] = np.append(self.data[ax]['Range'],event.xdata)

                # Replot all points
                self.set_segments(self.artists[ax]['Bounds'],self.data[ax]['Range'],ax)

                # Replot the integration
                if self.data[ax]['Integration']['Display']: self.integrate(ax)

                # No point is currently selected
                self.info['selected'] = False

        ##########################################
        ## INTEGRATE
        ##########################################

        # If in integrate mode
        elif self.info['interactive_mode'] == 'integrate':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return

            if not self.data[ax]['Integration']['Display']:
                # Plot the peaks
                self.integrate(ax)
                # Set display to true
                self.data[ax]['Integration']['Display'] = True
            else:
                # Clear the peaks
                self.artists[ax]['Integration'].set_paths([])
                # Set display to false
                self.data[ax]['Integration']['Display'] = False

        ##########################################
        ## FIND PEAKS
        ##########################################

        # If in find peaks mode
        elif self.info['interactive_mode'] == 'find_peaks':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return

            if not self.data[ax]['Peaks']['Display']:
                # Plot the peaks
                self.peaks(ax)
                # Set display to true
                self.data[ax]['Peaks']['Display'] = True
            else:
                # Clear the peaks
                self.set_segments(self.artists[ax]['Peaks'],[],ax)
                # Set display to false
                self.data[ax]['Peaks']['Display'] = False

        ##########################################
        ## CLEAR AXIS
        ##########################################

        # If in clear mode
        elif self.info['interactive_mode'] == 'clear':

            # If the artist is not meant for data display
            if ax in self.artists_global['Interactive Axes']:

                # Clear the data dictionary
                self.data[ax] = self.data_init()
                # Clear artists
                self.set_segments(self.artists[ax]['Bounds'],[],ax)
                self.set_segments(self.artists[ax]['New Bound'],[],ax)
                self.set_segments(self.artists[ax]['Selected Bound'],[],ax)

                # Clear the integration
                self.artists[ax]['Integration'].set_paths([])\

                # Clear the peaks
                self.set_segments(self.artists[ax]['Peaks'],[],ax)

                # Clear the legend
                ax.get_legend().remove()

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        # Update artists
        plt.show()

    ##########################################
    ## INTEGRATION METHODS
    ##########################################

    def integrate(self,ax):

        # Display the integration
        self.data[ax]['Integration']['Display'] = True

        # Clear the past integration
        self.artists[ax]['Integration'].set_paths([])

        # Determine the primary curve for the current axis
        curve = self.artists_global['Primary Artists'][ax]
        # Pull out the x and y data from that curve
        xdata,ydata = self.artists_global[ax][curve].get_data()
        # Determine the range of integration
        xmin,xmax = sorted(self.data[ax]['Range'])

        # Determine the indices of the data values within the range
        ii = np.squeeze(np.where((xdata > xmin) & (xdata < xmax)))
        # Redefine the baseline vars as their relevant subsets
        xdata,ydata = xdata[ii],ydata[ii]

        self.data[ax]['Integration']['xdata'] = xdata
        self.data[ax]['Integration']['ydata'] = ydata

        # Use scipy's simpsons integration technique -- negative for decreasing x range
        self.data[ax]['Integration']['area'] = area = simps(y=ydata,x=xdata)

        # Plot the integration
        self.artists[ax]['Integration'] = ax.fill_between(xdata,ydata,label=f'Area: {round(area,3)}',color='dodgerblue',alpha=0.6)
        ax.legend()

    ##########################################
    ## PEAKS METHODS
    ##########################################

    def peaks(self,ax):

        # Set the data
        height = self.data[ax]['height'] = self.info['height']
        dist = self.data[ax]['distance'] = self.info['distance']

        # Determine the primary curve for the current axis
        curve = self.artists_global['Primary Artists'][ax]
        xdata,ydata = self.artists_global[ax][curve].get_data()

        # Find the indeces of the data that are within the range
        xmin,xmax = sorted(self.data[ax]['Range'])
        ii = np.squeeze(np.where((xdata > xmin) & (xdata < xmax)))

        # Subset the x and y data
        xsub,ysub = xdata[ii],ydata[ii]

        # Find the peak indices
        ii = self.peak_vals(ysub,height,dist)

        # Save the xvalues of the peaks to the data dictionary
        self.data[ax]['Peaks']['xdata'] = xsub[ii]
        self.set_segments(self.artists[ax]['Peaks'],self.data[ax]['Peaks']['xdata'],ax)

    def peak_vals(self,spec,height,distance):
        normspec = spec/np.nanmax(spec)
        sp_medavg = medfilt(normspec,kernel_size=51)
        flat_spec = np.abs(normspec - sp_medavg)
        peaks,_ = find_peaks(flat_spec,height=height,distance=distance)
        return peaks



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 22 September 2022

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools
import pickle as pkl

from base import widget_base

class export(widget_base):
    def __init__(self,main_menu,menu=None,exp_path=None):
        '''
        '''
        self.widget_on = False
        self.main_menu = main_menu

        # Initialize defaults
        self.info = self.setup_defaults(exp_path)

        # Initialize all buttons
        self.button_list,self.toggle_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu

    def setup_defaults(self,exp_path):

        if exp_path is None: exp_path = ''
        info = {
        'exp_path':exp_path,
        'dpi':200
        }

        return info

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.export_button = widgets.Button(description='Export')

        def on_export_button_clicked(b, self = self):

            exp_dict = self.main_menu.export_dict()
            pkl.dump(exp_dict,open(f"{self.info['exp_path']}.pkl",'wb'))

        self.export_button.on_click(functools.partial(on_export_button_clicked, self=self))

        #####################

        # Add a button widget to save the figure
        self.save_button = widgets.Button(description='Save Plot')

        def on_save_button_clicked(b, self = self):

            plt.savefig(f"{self.info['exp_path']}.png",dpi=self.info['dpi'])

        self.save_button.on_click(functools.partial(on_save_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.export_path_text = widgets.Text(description='Path:',value=self.info['exp_path'],style={'description_width': 'initial'},layout=widgets.Layout(width='200px'))

        def export_path_entered(b, self = self):

            self.info['exp_path'] = self.export_path_text.value

        self.export_path_text.observe(functools.partial(export_path_entered, self=self))

        #####################

        # Add a button widget to add new points
        self.dpi_text = widgets.IntText(description='DPI:',value=self.info['dpi'],style={'description_width': 'initial'},layout=widgets.Layout(width='100px'))

        def dpi_entered(b, self = self):

            self.info['dpi'] = self.dpi_text.value

        self.dpi_text.observe(functools.partial(dpi_entered, self=self))

        #####################

        return [self.export_button,self.save_button,self.export_path_text,self.dpi_text],[]

    def update(self,artists,data): return



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 28 August 2022

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import copy as cp

from .fit_methods import fit_methods

class fit(fit_methods):
    def __init__(self,fit_module):

        # Tell the fit module that we are adding a gaussian fit
        fit_module.info['fit_dict']['--FIT NAME HERE--'] = self

        # Tell the fit module what parameters we want
        parameter_list = ['LIST PARAMS','HERE']

        # Set the model to be used
        self.model = self.NAMED_MODEL

        # Should there be a new axis added?
        self.new_ax = False

        # Setup the fit
        self.setup_fit(fit_module,parameter_list)

        return

    ##########################################
    ## THE MODEL
    ##########################################

    # MODEL TO BE REPLACED BY USER
    def NAMED_MODEL(self,xx,param1):
        # A function used for curve fitting
        y = xx
        return np.array(y)


    ##########################################
    ## OPTIONAL METHODS FOR EXTRA CONTROL
    ##########################################

    # ##########################################
    # ## PARAMETERS
    # ##########################################
    #
    # # Method to determine guess parameters and bounds
    # def get_params(self,parameters):
    #
    #     # Return the initial parameter guesses and bounds
    #     return(p0,bounds)
    #
    # ##########################################
    # ## PLOTTING METHODS
    # ##########################################
    #
    # # Method to add custom plotting for a certian fit
    # def plot_fit(self,fit_mod,stats,fit_number,ax,fit_ax=None):
    #
    #     return fit_line
    #
    # ##########################################
    # ## STATS
    # ##########################################
    #
    # # Method to add custom stats about a certain fit
    # def add_fit_stats(self,stats):
    #
    #     # Return a improved stats dict
    #     return(stats)
